"""Helpers for manman"""
import sys, os, time, glob
import subprocess

ConfigDir = '/operations/app_store/manman'

Verbose = 0

def printTime(): return time.strftime("%m%d:%H%M%S")
def printi(msg): print(f'inf_@{printTime()}: {msg}')
def printw(msg): print(f'WAR_@{printTime()}: {msg}')
def printe(msg): print(f'ERR_{printTime()}: {msg}')
def _printv(msg, level):
    if Verbose >= level:
        print(f'dbg{level}: {msg}')
def printv(msg): _printv(msg, 1)
def printvv(msg): _printv(msg, 2)




'''Definition of the second test apparatus.
'''
import os
homeDir = os.environ['HOME']

__version__ = 'v0.0.1 2025-05-21'#

# abbreviations:
help,cmd,process,cd = ['help','cmd','process','cd']

#``````````````````Properties, used by manman`````````````````````````````````
title = 'Test1 applications'

startup = {
'xclock':{help:'Analog xclock', 
  cmd:'xclock -analog'
  },
'top':{help:'Process viewer in separate xterm',
  cmd:'xterm top',
  },
'sleep10':{help:'Sleep for 10 seconds', 
  cmd:'sleep 10', process:'sleep 10'
  },
}



"""GUI for application deployment and monitoring of servers and 
applications related to specific apparatuses.
"""
__version__ = 'v1.0.0 2025-05-28'# added --condensed

import sys, os, argparse
from qtpy.QtWidgets import QApplication
from . import manman, helpers

#``````````````````Main```````````````````````````````````````````````````````
def main():
    global pargs
    parser = argparse.ArgumentParser('python -m manman',
      description=__doc__,
      formatter_class=argparse.ArgumentDefaultsHelpFormatter,
      epilog=f'Version {manman.__version__}')
    parser.add_argument('-c', '--configDir', help=\
      ('Root directory of config files, one config file per apparatus, '
      'if None, then ./config directory will be used'))
    parser.add_argument('-C', '--condensed', action='store_true', help=\
      'Condensed arrangement of tables: no headers, narrow columns')
    parser.add_argument('-t', '--interval', default=10., help=\
      'Interval in seconds of periodic checking. If 0 then no checking')
    parser.add_argument('-v', '--verbose', action='count', default=0, help=\
      'Show more log messages (-vv: show even more).')
    parser.add_argument('-z', '--zoomin', help=\
      'Zoom the application window by a factor, factor must be >= 1')
    parser.add_argument('apparatus', nargs='*', help=\
      ('Path of apparatus config files, can include wildcards. '
       'If None, then an interactive dialog will be opened to select files.')),
    pargs = parser.parse_args()
    helpers.Verbose = pargs.verbose
    if pargs.configDir is None and len(pargs.apparatus) == 0:
        pargs.configDir = 'config'
    manman.Window.pargs = pargs# transfer pargs to manman module

    # handle the --zoomin
    if pargs.zoomin is not None:
        os.environ["QT_SCALE_FACTOR"] = pargs.zoomin

    # arrange keyboard interrupt to kill the program
    import signal
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    
    # start GUI
    app = QApplication(sys.argv)
    window = manman.Window()
    window.show()
    app.exec_()
    print('Application exit')

if __name__ == '__main__':
    main()




"""Pypet for liteServer peak simulator with embedded plot
"""
__version__='v0.0.2 2025-04-15'# host="localhost;9701:"

_Namespace = "LITE"
host = "localhost;9701:"
dev = f"{host}dev1:"
server = f"{host}server:"
#``````````````````Definitions````````````````````````````````````````````````
# Python expressions and functions, used in the spreadsheet.
_=' '
def span(x,y): return {'span':[x,y]}
def color(*v): return {'color':v[0]} if len(v)==1 else {'color':list(v)}
pvplot = f"python3 -m pvplot -a L:{dev} x,y"

#``````````````````Page attributes, optional``````````````````````````````````
#_Page = {'editable':False, **color(252,252,237)}
_Page = {**color(240,240,240)}

_Columns = {
  1: {"justify": "center"},
  2: {"width": 100},
  3: {"justify": "right"},
  5: {"width": 400},
}

_Rows = [
['Performance:', {server+'perf':span(3,1)},_,_,{_:{'embed':pvplot,**span(1,10)}}],
["run", dev+"run", 'debug:', server+'debug'],
["status", {dev+"status":span(3,1)}],
["frequency", dev+"frequency", "nPoints:", dev+"nPoints"],
["background", {dev+"background":span(3,1)}],
["noise", dev+"noise", "swing:", dev+"swing"],
["peakPars", {dev+"peakPars":span(3,1)}],
#["x", {dev+"x":span(3,1)}],
#["y", {dev+"y":span(3,1)}],
['yMin:', dev+'yMin', 'yMax:', dev+'yMax'],
["rps", dev+"rps", "cycle:", dev+"cycle"],
[],
]



'''Definition of the second test apparatus.
'''
import os
homeDir = os.environ['HOME']

__version__ = 'v0.0.1 2025-05-21'#

# abbreviations:
help,cmd,process,cd = ['help','cmd','process','cd']

#``````````````````Properties, used by manman`````````````````````````````````
title = 'Test applications'

startup = {
'xclock':{help:'Digital xclock', 
  cmd:'xclock -digital'
  },
'htop':{help:'Process viewer in separate xterm',
  cmd:'xterm htop',
  },
'sleep30':{help:'Sleep for 30 seconds', 
  cmd:'sleep 30', process:'sleep 30'
  },
}



'''Definition of a test apparatus, running a liteServer with peak simulator.
The managers for this example could be installed with pip:
  pip install liteserver pvplot pypeto

The script should define dictionary startup.
Supported keys are:
  'cmd': command which will be used to start and stop the manager,
  'cd:   directory (if needed), from where to run the cmd,
  'process': used for stopping the manager using 'pkill -f', if cmd properly identifies the 
     manager, then this key is not necessary,
  'help': it will be used as a tooltip,
'''
import os
homeDir = os.environ['HOME']
epics_home = os.environ.get('EPICS_HOME')

__version__ = 'v0.1.5 2025-04-26'# added 'process' to better track of processes

# abbreviations:
help,cmd,process,cd = ['help','cmd','process','cd']

#``````````````````Properties, used by manman`````````````````````````````````
title = 'Peak Simulator'
startup = {
#       Operational managers
# liteServer-based
'peakSimulator':{help:
  'Lite server, simulating peaks and noise',
  cmd:		'python3 -m liteserver.device.litePeakSimulator -ilo -p9701',
  process:	'liteserver.device.litePeakSimulator -ilo -p9701',
  },
'plot it':{help:
  'Plotting tool for peakSimulator',
  cmd:		'python3 -m pvplot -aL:localhost;9701:dev1: x,y',
  process:	'pvplot -aL:localhost;9701:dev1: x,y',
  },
'control it':{help:
  'Automatic parameter editing tool of the peakSimulator',
  cmd:		'python3 -m pypeto -aLITE localhost;9701:dev1',
  process:	'pypeto -aLITE localhost;9701:dev1',
  },
'control&plot':{help:
  'Parameter editing with integrated plot',
  cmd:		'python3 -m pypeto -c config -f peakSimPlot',
  process:	'pypeto -c config -f peakSimPlot',
  #Note: It will look for config file: config/peakSimPlot_pp.py
  },
}
if epics_home is not None:
    startup.update({
# EPICS IOCs
'simScope':{help:
  'EPICS testAsynPortDriver, hosting a simulate oscilloscope',
  cd:f'{epics_home}/asyn/iocBoot/ioctestAsynPortDriver/',
  cmd:'screen -d -m -S simScope ../../bin/linux-x86_64/testAsynPortDriver st.cmd',
  process:'testAsynPortDriver st.cmd', 
},
#'tst_caproto_ioc':  {cmd:'python3 -m caproto.ioc_examples.simple --list-pvs',help:
#  'Simple IOC for testing EPICS Channel Access functionality'},
'pet_simScope':{help:
  'Parameter editing tool for simScope',
  cmd:  'python3 -m pypeto -f Controls/EPICS/simScope',
  },
})

#       Managers for testing and debugging
startup.update({
'tst_sleep30':{help:
  'sleep for 30 seconds', 
  cmd:'sleep 30', process:'sleep 30'
  },
})



"""GUI for application deployment and monitoring of servers and 
applications related to specific apparatus.
"""
__version__ = 'v1.0.3 2025-06-01'# Fixed: missing tooltips in Application column
#TODO: xdg_open does not launch if other editors not running. 

import sys, os, time, subprocess, argparse, threading
from functools import partial
from importlib import import_module

from qtpy import QtWidgets as QW, QtGui, QtCore

from . import helpers as H
from . import detachable_tabs

#``````````````````Constants``````````````````````````````````````````````````
ManCmds =       ['Check',    'Start',    'Stop',     'Command']
AllManActions = ['Check All','Start All','Stop All', 'Edit', 'Delete',
                'Condense', 'Uncondense']
Col = {'Applications':0, 'status':1, 'action':2, 'response':3}
BoldFont = QtGui.QFont("Helvetica", 14, QtGui.QFont.Bold)
FilePrefix = 'apparatus'

#``````````````````Helpers````````````````````````````````````````````````````
def select_files_interactively(directory, title=f'Select {FilePrefix}*.py files'):
    dialog = QW.QFileDialog()
    dialog.setFileMode( QW.QFileDialog.FileMode() )
    ffilter = f'pypet ({FilePrefix}*.py)'
    files = dialog.getOpenFileNames( None, title, directory, ffilter)[0]
    return files

def create_folderMap():
    # create map of {folder1: [file1,...], folder2...} from pargs.apparatus
    #print(f'c,a: {Window.pargs.configDir, Window.pargs.apparatus}')
    folders = {}
    if Window.pargs.configDir is None:
        files = [os.path.abspath(i) for i in Window.pargs.apparatus]
    else:
        absfolder = os.path.abspath(Window.pargs.configDir)
        if len(Window.pargs.apparatus) == 0:
            files = select_files_interactively(absfolder)
        else:
            files = [absfolder+'/'+i for i in Window.pargs.apparatus]
    for f in files:
        folder,tail = os.path.split(f)
        if not (tail.startswith(FilePrefix) and tail.endswith('.py')):
            H.printe(f'Config file should have prefix {FilePrefix} and suffix ".py"')
            sys.exit(1)
        if folder not in folders:
            folders[folder] = []
        folders[folder].append(tail)

    # sort the file lists
    for folder in folders:
        folders[folder].sort()
    return folders

def launch_default_editor(configFile):
    cmd = f'xdg-open {configFile}'
    H.printi(f'Launching editor: {cmd}')
    subprocess.call(cmd.split())

def is_process_running(cmdstart):
    try:
        subprocess.check_output(["pgrep", '-f', cmdstart])
        return True
    except subprocess.CalledProcessError:
        return False
#``````````````````Table Widget```````````````````````````````````````````````
def current_mytable():
    return Window.tabWidget.currentWidget()
class MyTable(QW.QTableWidget):
    def __init__(self, folder, fname, tabName):
        super().__init__()
        mname = fname[:-3]
        H.printv(f'importing {mname}')
        try:
            module = import_module(mname)
        except SyntaxError as e:
            H.printe(f'Syntax Error in {fname}: {e}')
            sys.exit(1)
        H.printv(f'imported {mname} {module.__version__}')
        self.startup = module.startup
        self.configFile = folder+'/'+fname
        self.setColumnCount(len(Col))
        self.setHorizontalHeaderLabels(Col.keys())
        self.manRow = {}
        try:
            H.printv(f'title: {module.title}')
            wideRow(self, 0, module.title)
        except:
            wideRow(self, 0,'Applications')

        # Create AllManActions combobox in top row
        sb = QW.QComboBox()
        sb.addItems(AllManActions)
        sb.activated.connect(self.tableWideAction)
        sb.setToolTip('Execute selected action for all applications of this tab')
        self.setCellWidget(0, Col['action'], sb)

        # Set up all rows 
        operationalManager = True
        for manName,props in self.startup.items():
            rowPosition = self.rowCount()
            if manName.startswith('tst_'):
                if operationalManager:
                    operationalManager = False
                    wideRow(self, rowPosition,'Test Apps')
                    rowPosition += 1
            insertRow(self, rowPosition)
            self.manRow[manName] = rowPosition
            item = QW.QTableWidgetItem(manName)
            item.setTextAlignment(QtCore.Qt.AlignCenter)
            try:    item.setToolTip(props['help'])
            except: pass
            self.setItem(rowPosition, Col['Applications'], item)
            if operationalManager:
                item.setFont(BoldFont)
                item.setBackground(QtGui.QColor('lightCyan'))
            self.setItem(rowPosition, Col['status'],
              QW.QTableWidgetItem('?'))
            sb = QW.QComboBox()
            sb.addItems(ManCmds)
            sb.activated.connect(partial(self.manAction, manName))
            try:    sb.setToolTip(f'Control of {manName}')
            except: pass
            self.setCellWidget(rowPosition, Col['action'], sb)
            self.setItem(rowPosition, Col['response'],
              QW.QTableWidgetItem(''))

        # Set up headers
        header = self.horizontalHeader()
        header.setStretchLastSection(True)
        if Window.pargs.condensed:
            self.set_headersVisibility(False)

    def manAction(self, manName:str, cmdIdx:int):
        # Execute action, indexed by cmdIdx
        try:
            cmd = ManCmds[cmdIdx]
        except Exception as e:
            H.printw(f'ManName,cmdIdx = {manName,cmdIdx}')
            return
        rowPosition = self.manRow[manName]
        H.printvv(f'manAction: {manName, cmd}')
        startup = self.startup
        cmdstart = startup[manName]['cmd']
        process = startup[manName].get('process', f'{cmdstart}')

        if cmd == 'Check':
            H.printvv(f'checking process {process} ')
            status = ['not running','is started'][is_process_running(process)]
            item = self.item(rowPosition,Col['status'])
            if not 'tst_' in manName:
                color = 'lightGreen' if 'started' in status else 'pink'
                item.setBackground(QtGui.QColor(color))
            item.setText(status)

        elif cmd == 'Start':
            self.item(rowPosition, Col['response']).setText('')
            if is_process_running(process):
                txt = f'Is already running manager {manName}'
                #print(txt)
                self.item(rowPosition, Col['response']).setText(txt)
                return
            H.printv(f'starting {manName}')
            item = self.item(rowPosition, Col['status'])
            if not 'tst_' in manName:
                item.setBackground(QtGui.QColor('lightYellow'))
            item.setText('starting...')
            path = startup[manName].get('cd')
            H.printi('Executing commands:')
            if path:
                path = path.strip()
                expandedPath = os.path.expanduser(path)
                try:
                    os.chdir(expandedPath)
                except Exception as e:
                    txt = f'ERR: in chdir: {e}'
                    self.item(rowPosition, Col['response']).setText(txt)
                    return
                print(f'cd {os.getcwd()}')
            print(cmdstart)
            expandedCmd = os.path.expanduser(cmdstart)
            cmdlist = expandedCmd.split()
            shell = startup[manName].get('shell',False)
            H.printv(f'popen: {cmdlist}, shell:{shell}')
            try:
                proc = subprocess.Popen(cmdlist, shell=shell, #close_fds=True,# env=my_env,
                  stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            except Exception as e:
                H.printv(f'Exception: {e}') 
                self.item(rowPosition, Col['response']).setText(str(e))
                return
            Window.timer.singleShot(5000,partial(self.deferredCheck,(manName,rowPosition)))

        elif cmd == 'Stop':
            self.item(rowPosition, Col['response']).setText('')
            H.printv(f'stopping {manName}')
            cmd = f'pkill -f "{process}"'
            H.printi(f'Executing:\n{cmd}')
            os.system(cmd)
            time.sleep(0.1)
            self.manAction(manName, ManCmds.index('Check'))

        elif cmd == 'Command':
            try:
                cd = startup[manName]['cd']
                cmd = f'cd {cd}; {cmdstart}'
            except Exception as e:
                cmd = cmdstart
            print(f'Command in row {rowPosition}:\n{cmd}')
            self.item(rowPosition, Col['response']).setText(cmd)
            return
        # Action was completed successfully, cleanup the status cell

    def set_headersVisibility(self, visible:bool):
        #print(f'set_headersVisibility {visible}')
        Window.pargs.condensed = False
        self.setColumnWidth(Col['action'], 10)
        self.horizontalHeader().setVisible(visible)
        self.verticalHeader().setVisible(visible)

    def tableWideAction(self, cmdidx:int):
        # Execute table-wide action
        #print(f'tableWideAction: {cmdidx}')
        if cmdidx == AllManActions.index('Edit'):
            launch_default_editor(self.configFile)
        elif cmdidx == AllManActions.index('Delete'):
            idx = Window.tabWidget.currentIndex()
            tabtext = Window.tabWidget.tabText(idx)
            H.printi(f'Deleting {idx,tabtext}')
            del Window.tableWidgets[tabtext]
            Window.tabWidget.removeTab(idx)
            self.deleteLater()# it is important to properly delete the associated widget
        elif cmdidx == AllManActions.index('Condense'):
            self.set_headersVisibility(False)
        elif cmdidx == AllManActions.index('Uncondense'):
            self.set_headersVisibility(True)
        else:
            for manName in self.startup:
                #print(f'man {manName,cmdidx}')
                if manName.startswith('tst') and cmdidx != ManCmds.index('Check'):
                    continue
                self.manAction(manName, cmdidx)

    def deferredCheck(self, args):
        manName,rowPosition = args
        self.manAction(manName, ManCmds.index('Check'))
        if 'start' not in self.item(rowPosition, Col['status']).text():
            self.item(rowPosition, Col['response']).setText('Failed to start')

#``````````````````Main Window````````````````````````````````````````````````
class Window(QW.QMainWindow):# it may sense to subclass it from QW.QMainWindow
    pargs = None
    tableWidgets = {}
    timer = QtCore.QTimer()

    def __init__(self):
        super().__init__()
        H.Verbose = Window.pargs.verbose
        folders = create_folderMap()
        if len(folders) == 0:
            sys.exit(1)
        H.printi(f'Configuration files: {folders}')
        self.setWindowTitle('manman')

        # Create tabWidget
        Window.tabWidget = detachable_tabs.DetachableTabWidget()
        Window.tabWidget.currentChanged.connect(periodicCheck)
        self.setCentralWidget(Window.tabWidget)
        H.printv(f'tabWidget created')

        # Add tables, configured from files, to tabs
        for folder,files in folders.items():
            sys.path.append(folder)
            for fname in files:
                tabName = fname[len(FilePrefix):-3]
                mytable = MyTable(folder, fname, tabName)
                Window.tableWidgets[tabName] = mytable
                #print(f'Adding tab: {fname}')
                Window.tabWidget.addTab(mytable, tabName)

        # Update tables and set up periodic check
        periodicCheck()
        if Window.pargs.interval != 0.:
            Window.timer.timeout.connect(periodicCheck)
            Window.timer.setInterval(int(Window.pargs.interval*1000.))
            Window.timer.start()

def wideRow(mytable, rowPosition, txt):
    insertRow(mytable, rowPosition)
    mytable.setSpan(rowPosition,0,1,2)
    item = QW.QTableWidgetItem(txt)
    item.setTextAlignment(QtCore.Qt.AlignCenter)
    item.setBackground(QtGui.QColor('lightGray'))
    item.setFont(BoldFont)
    mytable.setItem(rowPosition, Col['Applications'], item)

def insertRow(mytable, rowPosition):
    mytable.insertRow(rowPosition)
    mytable.setRowHeight(rowPosition, 1)  

def periodicCheck():
    # execute tableWideAction on current tab
    current_mytable().tableWideAction(ManCmds.index('Check'))
    # execute tableWideAction on all detached tabs
    for tabName,mytable in Window.tableWidgets.items():
        detached  = tabName in Window.tabWidget.detachedTabs
        #print(f'periodic for {tabName,detached}')
        if detached:
            mytable.tableWideAction(ManCmds.index('Check'))




"""GUI for application deployment and monitoring of servers and 
applications related to specific apparatus.
"""
__version__ = 'v0.5.2 2025-05-28'# Improved error handling

import sys, os, time, subprocess, argparse, threading
from functools import partial
from importlib import import_module

from qtpy import QtWidgets as QW, QtGui, QtCore

from . import helpers as H
from . import detachable_tabs

#``````````````````Constants``````````````````````````````````````````````````
ManCmds = ['Check','Start','Stop','Command']
Col = {'Applications':0, 'status':1, 'action':2, 'response':3}
AllManActions = ['Check All','Start All','Stop All', 'Edit', 'Delete']
BoldFont = QtGui.QFont("Helvetica", 14, QtGui.QFont.Bold)
FilePrefix = 'apparatus_'

#``````````````````Helpers````````````````````````````````````````````````````
def select_files_interactively(directory, title=f'Select {FilePrefix}*.py files'):
    dialog = QW.QFileDialog()
    dialog.setFileMode( QW.QFileDialog.FileMode() )
    ffilter = f'pypet ({FilePrefix}*.py)'
    files = dialog.getOpenFileNames( None, title, directory, ffilter)[0]
    return files

def create_folderMap():
    # create map of {folder1: [file1,...], folder2...} from pargs.apparatus
    #print(f'c,a: {Window.pargs.configDir, Window.pargs.apparatus}')
    folders = {}
    if Window.pargs.configDir is None:
        files = [os.path.abspath(i) for i in Window.pargs.apparatus]
    else:
        absfolder = os.path.abspath(Window.pargs.configDir)
        if len(Window.pargs.apparatus) == 0:
            files = select_files_interactively(absfolder)
        else:
            files = [absfolder+'/'+i for i in Window.pargs.apparatus]
    for f in files:
        folder,tail = os.path.split(f)
        if not (tail.startswith(FilePrefix) and tail.endswith('.py')):
            H.printe(f'Config file should have prefix {FilePrefix} and suffix ".py"')
            sys.exit(1)
        if folder not in folders:
            folders[folder] = []
        folders[folder].append(tail)
    return folders

def launch_default_editor(configFile):
    subprocess.call(['xdg-open', configFile])

def is_process_running(cmdstart):
    try:
        subprocess.check_output(["pgrep", '-f', cmdstart])
        return True
    except subprocess.CalledProcessError:
        return False
#``````````````````Table Widget```````````````````````````````````````````````
def current_mytable():
    return Window.tabWidget.currentWidget()
class MyTable(QW.QTableWidget):
    def __init__(self, startup, configFile):
        super().__init__()
        self.startup = startup
        self.configFile = configFile

    """ Help windows to adjust geometry
    def sizeHint(self):
        hh = self.horizontalHeader()
        vh = self.verticalHeader()
        fw = self.frameWidth() * 2
        return QtCore.QSize(
            hh.length() + vh.sizeHint().width() + fw,
            vh.length() + hh.sizeHint().height() + fw)
    """

#``````````````````Main Window````````````````````````````````````````````````
class Window(QW.QMainWindow):# it may sense to subclass it from QW.QMainWindow
    pargs = None
    tableWidgets = []
    manRow = {}
    #startup = None
    timer = QtCore.QTimer()

    def __init__(self):
        super().__init__()
        H.Verbose = Window.pargs.verbose
        folders = create_folderMap()
        if len(folders) == 0:
            sys.exit(1)
        H.printi(f'Configuration files: {folders}')

        # create tabWidget
        Window.tabWidget = detachable_tabs.DetachableTabWidget()
        Window.tabWidget.currentChanged.connect(periodicCheck)
        self.setCentralWidget(Window.tabWidget)
        H.printv(f'tabWidget created')

        for folder,files in folders.items():
            sys.path.append(folder)
            for fname in files:
                mytable = self.create_mytable(folder, fname)
                Window.tableWidgets.append(mytable)
                #print(f'Adding tab: {fname}')
                Window.tabWidget.addTab(mytable, fname[len(FilePrefix):-3])

        self.setWindowTitle('manman')
        #self.show()

        periodicCheck()
        if Window.pargs.interval != 0.:
            Window.timer.timeout.connect(periodicCheck)
            Window.timer.setInterval(int(Window.pargs.interval*1000.))
            Window.timer.start()

    def create_mytable(self, folder, fname):
        mname = fname[:-3]
        H.printv(f'importing {mname}')
        try:
            module = import_module(mname)
        except SyntaxError as e:
            H.printe(f'Syntax Error in {fname}: {e}')
            sys.exit(1)
        H.printv(f'imported {mname} {module.__version__}')
        startup = module.startup

        mytable =  MyTable(startup, folder+'/'+fname)
        #mytable.setWindowTitle('manman')
        mytable.setColumnCount(len(Col))
        mytable.setHorizontalHeaderLabels(Col.keys())
        wideRow(mytable, 0,'Operational Apps')
        
        sb = QW.QComboBox()
        sb.addItems(AllManActions)
        sb.activated.connect(allManAction)
        sb.setToolTip('Execute selected action for all applications')
        mytable.setCellWidget(0, Col['action'], sb)
        #return mytable

        operationalManager = True
        for manName in startup:
            rowPosition = mytable.rowCount()
            if manName.startswith('tst_'):
                if operationalManager:
                    operationalManager = False
                    wideRow(mytable, rowPosition,'Test Apps')
                    rowPosition += 1
            insertRow(mytable, rowPosition)
            self.manRow[manName] = rowPosition
            item = QW.QTableWidgetItem(manName)
            item.setTextAlignment(QtCore.Qt.AlignCenter)
            try:    item.setToolTip(startup[manName]['help'])
            except: pass
            mytable.setItem(rowPosition, Col['Applications'], item)
            if operationalManager:
                item.setFont(BoldFont)
                item.setBackground(QtGui.QColor('lightCyan'))
            mytable.setItem(rowPosition, Col['status'],
              QW.QTableWidgetItem('?'))
            sb = QW.QComboBox()
            sb.addItems(ManCmds)
            sb.activated.connect(partial(manAction,manName))
            try:    sb.setToolTip(startup[manName]['help'])
            except: pass
            mytable.setCellWidget(rowPosition, Col['action'], sb)
            mytable.setItem(rowPosition, Col['response'],
              QW.QTableWidgetItem(''))

        header = mytable.horizontalHeader()
        header.setStretchLastSection(True)

        return mytable

def wideRow(mytable, rowPosition,txt):
    insertRow(mytable, rowPosition)
    mytable.setSpan(rowPosition,0,1,2)
    item = QW.QTableWidgetItem(txt)
    item.setTextAlignment(QtCore.Qt.AlignCenter)
    item.setBackground(QtGui.QColor('lightGray'))
    item.setFont(BoldFont)
    mytable.setItem(rowPosition, Col['Applications'], item)

def insertRow(mytable, rowPosition):
    mytable.insertRow(rowPosition)
    mytable.setRowHeight(rowPosition, 1)  

def allManAction(cmdidx:int):
    #print(f'allManAction: {cmdidx}')
    mytable = current_mytable()
    if cmdidx == AllManActions.index('Edit'):
        launch_default_editor(mytable.configFile)
    elif cmdidx == AllManActions.index('Delete'):
        myTable = current_mytable()
        idx = Window.tabWidget.currentIndex()
        Window.tabWidget.removeTab(idx)
        myTable.deleteLater()# it is important to also properly delete the associated widget
    else:
        for manName in mytable.startup:
            if manName.startswith('tst') and cmdidx != ManCmds.index('Check'):
                continue
            manAction(manName, cmdidx)

def manAction(manName, cmdObj):
    # if called on click, then cmdObj is index in ManCmds, otherwise it is a string
    mytable = current_mytable()
    cmd = cmdObj if isinstance(cmdObj,str) else ManCmds[cmdObj]
    rowPosition = Window.manRow[manName]
    H.printv(f'manAction: {manName, cmd}')
    startup = current_mytable().startup
    cmdstart = startup[manName]['cmd']    
    process = startup[manName].get('process', f'{cmdstart}')

    if cmd == 'Check':
        H.printv(f'checking process {process} ')
        status = ['not running','is started'][is_process_running(process)]
        item = mytable.item(rowPosition,Col['status'])
        if not 'tst_' in manName:
            color = 'lightGreen' if 'started' in status else 'pink'
            item.setBackground(QtGui.QColor(color))
        item.setText(status)
            
    elif cmd == 'Start':
        mytable.item(rowPosition, Col['response']).setText('')
        if is_process_running(process):
            txt = f'Is already running manager {manName}'
            #print(txt)
            mytable.item(rowPosition, Col['response']).setText(txt)
            return
        H.printv(f'starting {manName}')
        item = mytable.item(rowPosition, Col['status'])
        if not 'tst_' in manName:
            item.setBackground(QtGui.QColor('lightYellow'))
        item.setText('starting...')
        path = startup[manName].get('cd')
        H.printi('Executing commands:')
        if path:
            path = path.strip()
            expandedPath = os.path.expanduser(path)
            try:
                os.chdir(expandedPath)
            except Exception as e:
                txt = f'ERR: in chdir: {e}'
                mytable.item(rowPosition, Col['response']).setText(txt)
                return
            print(f'cd {os.getcwd()}')
        print(cmdstart)
        expandedCmd = os.path.expanduser(cmdstart)
        cmdlist = expandedCmd.split()
        shell = startup[manName].get('shell',False)
        H.printv(f'popen: {cmdlist}, shell:{shell}')
        try:
            proc = subprocess.Popen(cmdlist, shell=shell, #close_fds=True,# env=my_env,
              stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        except Exception as e:
            H.printv(f'Exception: {e}') 
            mytable.item(rowPosition, Col['response']).setText(str(e))
            return
        Window.timer.singleShot(5000,partial(deferredCheck,(manName,rowPosition)))

    elif cmd == 'Stop':
        mytable.item(rowPosition, Col['response']).setText('')
        H.printv(f'stopping {manName}')
        cmd = f'pkill -f "{process}"'
        H.printi(f'Executing:\n{cmd}')
        os.system(cmd)
        time.sleep(0.1)
        manAction(manName, 'Check')

    elif cmd == 'Command':
        try:
            cd = startup[manName]['cd']
            cmd = f'cd {cd}; {cmdstart}'
        except Exception as e:
            cmd = cmdstart
        print(f'Command:\n{cmd}')
        mytable.item(rowPosition, Col['response']).setText(cmd)
        return
    # Action was completed successfully, cleanup the status cell

def deferredCheck(args):
    manName,rowPosition = args
    manAction(manName, 'Check')
    mytable = current_mytable()
    if 'start' not in mytable.item(rowPosition, Col['status']).text():
        mytable.item(rowPosition, Col['response']).setText('Failed to start')

def periodicCheck():
    allManAction('Check')




'''Command line tool to start servers and managers
'''
__version__ = 'v0.1.4 2024-10-27'#
import sys, os, time, subprocess, argparse, threading
from functools import partial
from importlib import import_module

from . import helpers as H

Apparatus = H.list_of_apparatus()

ManCmds = ['Check','Start','Stop','Command']

def manAction(manName, cmd):
    H.printv(f'manAction: {manName, cmd}')
    cmdstart = Startup[manName]['cmd']    
    if cmd == 'Check':
        H.printv(f'checking process {cmdstart} ')
        if H.is_process_running(cmdstart):
            print(f'Manager "{manName}" \tstarted')#, process name: "{cmdstart}"')
            return os.EX_OK
        else:
            print(f'Manager "{manName}" \tis not running')
            return os.EX_UNAVAILABLE
            
    elif cmd == 'Start':
        H.printv(f'starting {manName}')
        if H.is_process_running(cmdstart):
            H.printe(f'Manager "{manName}" is already running.')
            return os.EX_CANTCREAT

        cmdlist = cmdstart.split()
        H.printv(f'popen: {cmdlist}')
        try:
            process = subprocess.Popen(cmdlist, #close_fds=True,# env=my_env,
              stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        except Exception as e:
            H.printe(f'Exception starting {manName}: {e}') 
            return os.EX_IOERR

        time.sleep(5)# 3 is small for PLC device
        H.printv('slept 5 seconds')
        ex = manAction(manName, 'Check')
        if ex != os.EX_OK:
            return os.EX_UNAVAILABLE

    elif cmd == 'Stop':
        H.printv(f'stopping {manName}')
        process = Startup[manName].get('process', f'{cmdstart}')
        cmd = f'pkill -f "{process}"'
        H.printv(f'executing: {cmd}')
        os.system(cmd)
        time.sleep(0.1)
        ex = manAction(manName, 'Check')
        if ex != os.EX_UNAVAILABLE:
            return os.EX_SOFTWARE

    elif cmd == 'Command':
        try:
            cd = Startup[manName]['cd']
            cmd = f'cd {cd}; {cmdstart}'
        except Exception as e:
            cmd = cmdstart
        print(f'Start command for "{manName}": "{cmd}"')
        return os.EX_OK

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description=__doc__,
      formatter_class=argparse.ArgumentDefaultsHelpFormatter,
      epilog=f'Version {__version__}')
    parser.add_argument('-a','--apparatus', choices=Apparatus, default='TST', help='Which apparatus to control')
    parser.add_argument('-c', '--configDir', default=H.ConfigDir, help=\
      'Directory, containing apparatus configuration scripts')
    parser.add_argument('-m', '--manager', default='all',
help='Apply command to a particular manager (or all) of the apparatus')
    parser.add_argument('-t', '--test', action='store_true',
help='Include non-operational (test) managers') 
    parser.add_argument('-v', '--verbose', action='count', default=0, help=\
      'Show more log messages (-vv: show even more).')
    parser.add_argument('command', nargs='?', choices=ManCmds, default='Check')
    pargs = parser.parse_args()
    H.printv(f'pargs: {pargs}')

    # import the manager
    mname = 'manman.apparatus_'+pargs.apparatus
    module = import_module(mname)
    print(f'imported {mname} {module.__version__}')
    Startup = module.startup
    mname = 'manman.apparatus_'+pargs.apparatus
    module = import_module(mname)

    if pargs.manager == 'all':
        pargs.manager = list(Startup.keys())
    else:
        pargs.manager = [pargs.manager]
    H.printv(f'Managers: {pargs.manager}')

    for manName in pargs.manager:
        if manName not in Startup:
            H.printe(f'Wrong manager, supported are: {",".join(Startup.keys())}')
            sys.exit(os.EX_USAGE)
        if not pargs.test:
            if manName.startswith('tst_'):
                continue
        manAction(manName, pargs.command)




"""GUI for application deployment and monitoring of servers and 
applications related to specific apparatus.
"""
__version__ = 'v0.4.0 2025-05-21'# 
#TODO: Use QTableView instead of QTableWidget, it is more flexible

import sys, os, time, subprocess, argparse, threading
from functools import partial
from importlib import import_module

from PyQt5 import QtWidgets as QW, QtGui, QtCore

from . import helpers as H
from . import detachable_tabs

Apparatus = H.list_of_apparatus()

ManCmds = ['Check','Start','Stop','Command']
Col = {'Managers':0, 'status':1, 'action':2, 'response':3}
BoldFont = QtGui.QFont("Helvetica", 14, QtGui.QFont.Bold)
LastColumnWidth=400

qApp = QW.QApplication(sys.argv)

class MyTable(QW.QTableWidget):

    def sizeHint(self):
        hh = self.horizontalHeader()
        vh = self.verticalHeader()
        fw = self.frameWidth() * 2
        return QtCore.QSize(
            hh.length() + vh.sizeHint().width() + fw,
            vh.length() + hh.sizeHint().height() + fw)

def currentDaTable():
    return MyWin.tabWidget.currentWidget().mytable

class MyWin(QW.QMainWindow):# it may sense to subclass it from QW.QMainWindow
    tw = None
    tabWidget = None
    tableWidgets = []
    manRow = {}
    startup = None
    timer = QtCore.QTimer()
    firstAction=True

    def __init__(self):
        QW.QWidget.__init__(self)# is it needed?
        MyWin.tabWidget = detachable_tabs.DetachableTabWidget()
        self.setCentralWidget(MyWin.tabWidget)
        print(f'tabWidget created')

        MyWin.tw = self.create_mytable('')

        if pargs.interval != 0.:
            MyWin.timer.timeout.connect(periodicCheck)
            MyWin.timer.setInterval(int(pargs.interval*1000.))
            MyWin.timer.start()
        MyWin.tw.show()

    def create_mytable(self, filename):
        tw =  MyTable()
        tw.setWindowTitle('manman')
        tw.setColumnCount(4)
        tw.setHorizontalHeaderLabels(Col.keys())
        wideRow(tw, 0,'Operational Apps')
        
        sb = QW.QComboBox()
        sb.addItems(['Check All','Start All','Stop All', 'Edit '])
        sb.activated.connect(allManAction)
        sb.setToolTip('Execute selected action for all applications')
        tw.setCellWidget(0, Col['action'], sb)

        operationalManager = True
        for manName in MyWin.startup:
            rowPosition = tw.rowCount()
            if manName.startswith('tst_'):
                if operationalManager:
                    operationalManager = False
                    wideRow(tw, rowPosition,'Test Apps')
                    
                    rowPosition += 1
            insertRow(tw, rowPosition)
            self.manRow[manName] = rowPosition
            item = QW.QTableWidgetItem(manName)
            item.setTextAlignment(QtCore.Qt.AlignCenter)
            try:    item.setToolTip(MyWin.startup[manName]['help'])
            except: pass
            tw.setItem(rowPosition, Col['Managers'], item)
            if operationalManager:
                item.setFont(BoldFont)
                item.setBackground(QtGui.QColor('lightCyan'))
            tw.setItem(rowPosition, Col['status'],
              QW.QTableWidgetItem('?'))
            sb = QW.QComboBox()
            sb.addItems(ManCmds)
            sb.activated.connect(partial(manAction,manName))
            try:    sb.setToolTip(MyWin.startup[manName]['help'])
            except: pass
            tw.setCellWidget(rowPosition, Col['action'], sb)
            tw.setItem(rowPosition, Col['response'],
              QW.QTableWidgetItem(''))
        return tw

def wideRow(tw, rowPosition,txt):
    insertRow(tw, rowPosition)
    tw.setSpan(rowPosition,0,1,2)
    item = QW.QTableWidgetItem(txt)
    item.setTextAlignment(QtCore.Qt.AlignCenter)
    item.setBackground(QtGui.QColor('lightGray'))
    item.setFont(BoldFont)
    tw.setItem(rowPosition, Col['Managers'], item)

def insertRow(tw, rowPosition):
    tw.insertRow(rowPosition)
    tw.setRowHeight(rowPosition, 1)  

def allManAction(cmdidx:int):
    H.printv(f'allManAction: {cmdidx}')
    if cmdidx == 3:
        _edit()
        return
    for manName in MyWin.startup:
        #if manName.startswith('tst'):
        #    continue
        manAction(manName, cmdidx)

def _edit():
    subprocess.call(['xdg-open', pargs.configFile])

def manAction(manName, cmdObj):
    # if called on click, then cmdObj is index in ManCmds, otherwise it is a string
    if MyWin.firstAction:
        MyWin.tw.setColumnWidth(3, LastColumnWidth)
        MyWin.firstAction = False
    cmd = cmdObj if isinstance(cmdObj,str) else ManCmds[cmdObj]
    rowPosition = MyWin.manRow[manName]
    H.printv(f'manAction: {manName, cmd}')
    cmdstart = MyWin.startup[manName]['cmd']    
    process = MyWin.startup[manName].get('process', f'{cmdstart}')

    if cmd == 'Check':
        H.printv(f'checking process {process} ')
        status = ['not running','is started'][H.is_process_running(process)]
        item = MyWin.tw.item(rowPosition,Col['status'])
        if not 'tst_' in manName:
            color = 'lightGreen' if 'started' in status else 'pink'
            item.setBackground(QtGui.QColor(color))
        item.setText(status)
            
    elif cmd == 'Start':
        MyWin.tw.item(rowPosition, Col['response']).setText('')
        if H.is_process_running(process):
            txt = f'Is already running manager {manName}'
            #print(txt)
            MyWin.tw.item(rowPosition, Col['response']).setText(txt)
            return
        H.printv(f'starting {manName}')
        item = MyWin.tw.item(rowPosition, Col['status'])
        if not 'tst_' in manName:
            item.setBackground(QtGui.QColor('lightYellow'))
        item.setText('starting...')
        path = MyWin.startup[manName].get('cd')
        H.printi('Executing commands:')
        if path:
            path = path.strip()
            expandedPath = os.path.expanduser(path)
            try:
                os.chdir(expandedPath)
            except Exception as e:
                txt = f'ERR: in chdir: {e}'
                MyWin.tw.item(rowPosition, Col['response']).setText(txt)
                return
            print(f'cd {os.getcwd()}')
        print(cmdstart)
        expandedCmd = os.path.expanduser(cmdstart)
        cmdlist = expandedCmd.split()
        shell = MyWin.startup[manName].get('shell',False)
        H.printv(f'popen: {cmdlist}, shell:{shell}')
        try:
            proc = subprocess.Popen(cmdlist, shell=shell, #close_fds=True,# env=my_env,
              stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        except Exception as e:
            H.printv(f'Exception: {e}') 
            MyWin.tw.item(rowPosition, Col['response']).setText(str(e))
            return
        MyWin.timer.singleShot(5000,partial(deferredCheck,(manName,rowPosition)))

    elif cmd == 'Stop':
        MyWin.tw.item(rowPosition, Col['response']).setText('')
        H.printv(f'stopping {manName}')
        cmd = f'pkill -f "{process}"'
        H.printi(f'Executing:\n{cmd}')
        os.system(cmd)
        time.sleep(0.1)
        manAction(manName, 'Check')

    elif cmd == 'Command':
        try:
            cd = MyWin.startup[manName]['cd']
            cmd = f'cd {cd}; {cmdstart}'
        except Exception as e:
            cmd = cmdstart
        print(f'Command:\n{cmd}')
        MyWin.tw.item(rowPosition, Col['response']).setText(cmd)
        return
    # Action was completed successfully, cleanup the status cell

def deferredCheck(args):
    manName,rowPosition = args
    manAction(manName, 'Check')
    if 'start' not in MyWin.tw.item(rowPosition, Col['status']).text():
        MyWin.tw.item(rowPosition, Col['response']).setText('Failed to start')

def periodicCheck():
    allManAction('Check')

def main():
    global pargs
    parser = argparse.ArgumentParser('python -m manman',
      description=__doc__,
      formatter_class=argparse.ArgumentDefaultsHelpFormatter,
      epilog=f'Version {__version__}')
    parser.add_argument('-c', '--configDir', default=H.ConfigDir, help=\
      'Directory, containing apparatus configuration scripts')
    parser.add_argument('-t', '--interval', default=10., help=\
      'Interval in seconds of periodic checking. If 0 then no checking')
    parser.add_argument('-v', '--verbose', action='count', default=0, help=\
      'Show more log messages (-vv: show even more).')
    parser.add_argument('apparatus', help=\
      'Apparatus', nargs='?', choices=Apparatus, default='TST')
    pargs = parser.parse_args()
    #pargs.log = None# disable logging fo now
    H.Constant.verbose = pargs.verbose

    mname = 'apparatus_'+pargs.apparatus
    pargs.configFile = f'{pargs.configDir}/{mname}.py'
    print(f'Config file: {pargs.configFile}')
    module = import_module(mname)
    #print(f'imported {mname} {module.__version__}')
    MyWin.startup = module.startup

    MyWin()
    allManAction('Check')

    # arrange keyboard interrupt to kill the program
    import signal
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    
    #start GUI
    try:
        qApp.instance().exec_()
        #sys.exit(qApp.exec_())
    except Exception as e:#KeyboardInterrupt:
        # This exception never happens
        print('keyboard interrupt: exiting')
    print('Application exit')

if __name__ == '__main__':
    main()

